//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#ifndef FACTOR_ANGULAR_MOMENTUM_H
#define FACTOR_ANGULAR_MOMENTUM_H

#include "bodydynamics/math/force_torque_inertial_delta_tools.h"
#include <core/feature/feature_motion.h>
#include <core/factor/factor_autodiff.h>

namespace wolf
{
using namespace Eigen;
using namespace bodydynamics;

WOLF_PTR_TYPEDEFS(FactorAngularMomentum);

/**
 * @brief
 *
 * This factor evaluates the real angular velocity against the one obtained with the pre-integrated angular momentum
 * and the moment of inertia.
 *
 * State blocks considered:
 *  - Frame
 *    'L' angular momentum
 *  - Capture
 *    'I' IMU biases
 *  - Sensor Force Torque Inertial
 *    'i' inertia matrix components (we are assuming that it is a diagonal matrix)
 *
 * The residual computed has the following components, in this order
 *  - angular velocity error
 */
class FactorAngularMomentum : public FactorAutodiff<FactorAngularMomentum, 3, 3, 6, 3, 4>  // State Blocks: L, I, i, q
{
  public:
    FactorAngularMomentum(const FeatureBasePtr&   _ftr,        // gets measurement and access to parents
                          const ProcessorBasePtr& _processor,  // gets access to processor and sensor
                          bool                    _apply_loss_function,
                          FactorStatus            _status = FAC_ACTIVE);

    ~FactorAngularMomentum() override = default;

    template <typename T>
    bool operator()(const T* const _L,  // ang momentum
                    const T* const _I,  // IMU bias (acc and gyro)
                    const T* const _i,  // inertia matrix
                    const T* const _q,  // quaternion
                    T*             _res) const;     // residual

    template <typename D1, typename D2, typename D3, typename D4, typename D5>
    bool residual(const MatrixBase<D1>& _L,
                  const MatrixBase<D2>& _I,
                  const MatrixBase<D3>& _i,
                  const QuaternionBase<D4>& _q,
                  MatrixBase<D5>&       _res) const;

    Vector3d residual() const;

  private:
    Vector3d measurement_ang_vel_;  // gyroscope measurement data
    Matrix3d sqrt_info_upper_;
};

inline FactorAngularMomentum::FactorAngularMomentum(const FeatureBasePtr&   _ftr,
                                                    const ProcessorBasePtr& _processor,
                                                    bool                    _apply_loss_function,
                                                    FactorStatus            _status)
    : FactorAutodiff<FactorAngularMomentum, 3, 3, 6, 3, 4>(
          "FactorAngularMomentum",
          TOP_MOTION,
          _ftr,        // parent feature
          nullptr,     // frame other
          nullptr,     // capture other
          nullptr,     // feature other
          nullptr,     // landmark other
          _processor,  // processor
          _apply_loss_function,
          _status,
          _ftr->getFrame()->getStateBlock('L'),         // current linear momentum
          _ftr->getCapture()->getStateBlock('I'),       // IMU bias
          _processor->getSensor()->getStateBlock('i'),  // moment of inertia
          _ftr->getFrame()->getStateBlock('O')),        // quaternion
      measurement_ang_vel_(_ftr->getMeasurement()),
      sqrt_info_upper_(_ftr->getMeasurementSquareRootInformationUpper())  // Buscar funcio correcte
{
    //
}

template <typename D1, typename D2, typename D3, typename D4, typename D5>
inline bool FactorAngularMomentum::residual(const MatrixBase<D1>& _L,
                                            const MatrixBase<D2>& _I,
                                            const MatrixBase<D3>& _i,
                                            const QuaternionBase<D4>& _q,
                                            MatrixBase<D5>&       _res) const
{
    MatrixSizeCheck<3, 1>::check(_res);

    /* Will do the following:
     *
     * Compute the real angular velocity through the measurement and the IMU bias
     *    w_real = w_m - w_b
     *
     * Compute the angular velocity obtained by the relation between L pre-integrated and the i
     *    w_est = i^(-1)*q.conjugate()*L
     *
     * Compute error between them
     *    err = w_m - w_b - i^(-1)*q.conjugate()*L
     *
     * Compute residual
     *    res = W * err , with W the sqrt of the info matrix.
     */

    typedef typename D5::Scalar T;

    Matrix<T, 3, 1> w_real = measurement_ang_vel_ - _I.template segment<3>(3);
    Matrix<T, 3, 1> L_local;
    L_local = _q.conjugate()*_L;                //we transform que _L from the global frame to the local frame
    const auto&     Lx     = L_local(0);
    const auto&     Ly     = L_local(1);
    const auto&     Lz     = L_local(2);
    const auto&     ixx    = _i(0);
    const auto&     iyy    = _i(1);
    const auto&     izz    = _i(2);
    Matrix<T, 3, 1> w_est(Lx / ixx, Ly / iyy, Lz / izz);
    Matrix<T, 3, 1> err = w_real - w_est;
    _res                = sqrt_info_upper_ * err;

    return true;
}

inline Vector3d FactorAngularMomentum::residual() const
{
    Vector3d res(3);
    Vector3d L = getFrame()->getStateBlock('L')->getState();
    Vector6d I = getCapture()->getStateBlock('I')->getState();
    Vector3d i = getSensor()->getStateBlock('i')->getState();
    Quaterniond q;
    q.coeffs() = getFrame()->getStateBlock('O')->getState();

    residual(L, I, i, q, res);
    return res;
}

template <typename T>
inline bool FactorAngularMomentum::operator()(const T* const _L, const T* const _I, const T* const _i, const T* const _q, T* _res) const
{
    Map<const Matrix<T, 3, 1>> L(_L);
    Map<const Matrix<T, 6, 1>> I(_I);
    Map<const Matrix<T, 3, 1>> i(_i);
    Map<const Quaternion<T>> q(_q);
    Map<Matrix<T, 3, 1>>       res(_res);

    residual(L, I, i, q, res);

    return true;
}

}  // namespace wolf

#endif  // FACTOR_ANGULAR_MOMENTUM_H
