//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file processor_inertial_kinematics.cpp
 *
 *  Created on: Feb 27, 2020
 *      \author: jsola
 */

#include "bodydynamics/processor/processor_inertial_kinematics.h"


namespace wolf{


inline ProcessorInertialKinematics::ProcessorInertialKinematics(ParamsProcessorInertialKinematicsPtr _params_ikin) :
        ProcessorBase("ProcessorInertialKinematics", 3, _params_ikin),
        params_ikin_(std::make_shared<ParamsProcessorInertialKinematics>(*_params_ikin)),
        cap_origin_ptr_(nullptr)
{
}

void ProcessorInertialKinematics::configure(SensorBasePtr _sensor)
{
}


inline void ProcessorInertialKinematics::processCapture(CaptureBasePtr _capture)
{
    if (!_capture)
    {
        WOLF_ERROR("Received capture is nullptr.");
        return;
    }

    // nothing to do if any of the two buffer is empty
    if(buffer_frame_.empty()){
        WOLF_DEBUG("PInertialKinematic: KF pack buffer empty, time ",  _capture->getTimeStamp());
        return;
    }
    if(buffer_frame_.empty()){
        WOLF_DEBUG("PInertialKinematics: Capture buffer empty, time ",  _capture->getTimeStamp());
        return;
    }

    // PROBLEM with this method:
    // if the incoming capture _capture matches one of the stored KF it will not be used to create
    // a factor until the next one arrives since it is not in the capture buffer yet (storeCapture is
    // done AFTER processCapture)

    // 1. get corresponding KF
    auto buffer_frame_it = buffer_frame_.getContainer().begin();
    auto buffer_capture_it = buffer_capture_.getContainer().begin();

    auto sensor_angvel = getProblem()->findSensor(params_ikin_->sensor_angvel_name);
    while ((buffer_frame_it != buffer_frame_.getContainer().end())
        && (buffer_capture_it != buffer_capture_.getContainer().end()))
    {

        bool time_ok = checkTimeTolerance(buffer_frame_it->first, buffer_capture_it->first);
        if (time_ok) {
            CaptureBasePtr cap_angvel = buffer_frame_it->second->getCaptureOf(sensor_angvel);
            auto min_ts = (buffer_frame_it->first < buffer_capture_it->first) ? buffer_frame_it->first : buffer_capture_it->first;
            if (cap_angvel && cap_angvel->getStateBlock('I')){  // TODO: or only cap_angvel?
                // cast incoming capture to the InertialKinematics type, add it to the keyframe
                auto kf = buffer_frame_it->second;
                auto cap_ikin = std::static_pointer_cast<CaptureInertialKinematics>(buffer_capture_it->second);
                cap_ikin->link(kf);
                createInertialKinematicsFactor(cap_ikin,
                                               cap_angvel,
                                               cap_origin_ptr_);
                // update pointer to origin capture (the previous one attached to a KF) if we have created a new factor
                cap_origin_ptr_ = buffer_capture_it->second;
                buffer_capture_it++;
                buffer_frame_it++;
            }
            else {
                // if time ok but no capture angvel yet, there is not gonna be any in the next KF of the buffer
                break;
                buffer_capture_it++;
                buffer_frame_it++;
            }

            ////////////////
            // remove everything before (Inclusive if equal) this timestamp -> the cap_angvel is yet to come
            buffer_frame_.removeUpTo(min_ts);
            buffer_capture_.removeUpTo(min_ts);
        }
        else {
            // if a time difference between captures and KF pack, we increment the oldest iterator
            if (buffer_capture_it->first < buffer_frame_it->first){
                buffer_capture_it++;
            }
            else {
                buffer_frame_it++;
            }
        }
    }
}


inline bool ProcessorInertialKinematics::createInertialKinematicsFactor(CaptureInertialKinematicsPtr _cap_ikin, CaptureBasePtr _cap_angvel, CaptureBasePtr _cap_ikin_origin)
{
    // 3. explore all observations in the capture and create feature
    Eigen::Vector9d data = _cap_ikin->getData();
    Eigen::Matrix3d B_I_q = _cap_ikin->getBIq();
    Eigen::Vector3d B_Lc_m = _cap_ikin->getBLcm();
    auto sensor_ikin = std::static_pointer_cast<SensorInertialKinematics>(getSensor());
    auto sensor_angvel_base = getProblem()->findSensor(params_ikin_->sensor_angvel_name);
    Eigen::Matrix3d Qp = pow(sensor_ikin->getPbcNoiseStd(), 2) * Eigen::Matrix3d::Identity();
    Eigen::Matrix3d Qv = pow(sensor_ikin->getVbcNoiseStd(), 2) * Eigen::Matrix3d::Identity();
    Eigen::Matrix3d Qw = sensor_angvel_base->getNoiseCov().bottomRightCorner<3,3>();
    Eigen::Matrix9d Q_ikin_err = computeIKinCov(Qp, Qv, Qw,
                                                data.head<3>() - _cap_ikin->getSensorIntrinsic()->getState(),   // bias value of the capture that was just created
                                                data.tail<3>() - _cap_angvel->getSensorIntrinsic()->getState().tail<3>(),
                                                B_I_q);
    auto feat_ikin = FeatureBase::emplace<FeatureInertialKinematics>(_cap_ikin, data, Q_ikin_err, B_I_q, B_Lc_m);

    // 4. Create inertial kinematics factor
    auto fac = FactorBase::emplace<FactorInertialKinematics>(feat_ikin,
                            feat_ikin,
                            _cap_angvel->getSensorIntrinsic(),
                            shared_from_this(),
                            false);

    // 5. Create bias drift factor if an origin capture exists -> zero difference + covariance as integration of a Brownian noise
    if (_cap_ikin_origin){
        double dt_drift = _cap_ikin->getTimeStamp() - _cap_ikin_origin->getTimeStamp();
        Eigen::Matrix3d cov_drift = Eigen::Matrix3d::Identity() * pow(params_ikin_->std_bp_drift, 2) * dt_drift;
        FeatureBasePtr feat_diff = FeatureBase::emplace<FeatureBase>(_cap_ikin, "ComBiasDrift", Eigen::Vector3d::Zero(), cov_drift, 
                                                                     FeatureBase::UncertaintyType::UNCERTAINTY_IS_COVARIANCE);
        StateBlockPtr sb_pbc1 = _cap_ikin_origin->getSensorIntrinsic();
        StateBlockPtr sb_pbc2 = _cap_ikin->getSensorIntrinsic();
        FactorBlockDifferencePtr Fac = FactorBase::emplace<FactorBlockDifference>(
                feat_diff, feat_diff, sb_pbc1, sb_pbc2, nullptr, _cap_ikin_origin
        );
    }

    return true;
}



//////////////////////////
// Covariance computations
//////////////////////////

Eigen::Matrix9d computeIKinJac(const Eigen::Vector3d& b_p_bc_unbiased,
                               const Eigen::Vector3d& b_omg_b_unbiased,
                               const Eigen::Matrix3d& Iq)
{
    Eigen::Matrix9d J; J.setZero();

    Eigen::Matrix3d wx = skew(b_omg_b_unbiased);
    Eigen::Matrix3d px = skew(b_p_bc_unbiased);
    // Starting from top left, to the right and then next row
    J.topLeftCorner<3,3>() = Eigen::Matrix3d::Identity();
    // J.block<3,3>(0,3) = Matrix3d::Zero();
    // J.topRightCorner<3,3>() = Matrix3d::Zero();
    J.block<3,3>(3,0) = - wx;
    J.block<3,3>(3,3) = - Eigen::Matrix3d::Identity();
    J.block<3,3>(3,6) =   px;
    // J.bottomLeftCorner<3,3>() = Matrix3d::Zero();
    // J.block<3,3>(6,3) = Matrix3d::Zero();
    J.bottomRightCorner<3,3>() = -Iq;

    return J;
}


Eigen::Matrix9d computeIKinCov(const Eigen::Matrix3d& Qp,
                               const Eigen::Matrix3d& Qv,
                               const Eigen::Matrix3d& Qw,
                               const Eigen::Vector3d& b_p_bc_unbiased,
                               const Eigen::Vector3d& b_omg_b_unbiased,
                               const Eigen::Matrix3d& Iq)
{
    Eigen::Matrix9d cov; cov.setZero();

    Eigen::Matrix3d px = skew(b_p_bc_unbiased);
    Eigen::Matrix3d wx = skew(b_omg_b_unbiased);
    // Starting from top left, to the right and then next row
    cov.topLeftCorner<3,3>() = Qp;
    cov.block<3,3>(0,3) = Qp * wx;
    // cov.topRightCorner<3,3>() = Matrix3d::Zero();
    cov.block<3,3>(3,0) = cov.block<3,3>(0,3).transpose();
    cov.block<3,3>(3,3) = -wx*Qp*wx + Qv - px*Qw*px;
    cov.block<3,3>(3,6) = -px*Qw*Iq;
    // cov.bottomLeftCorner<3,3>() = Matrix3d::Zero();
    cov.block<3,3>(6,3) = Iq*Qw*px;
    cov.bottomRightCorner<3,3>() = Iq*Qw*Iq;

    return cov;
}

} /* namespace wolf */

// Register in the FactoryProcessor
#include "core/processor/factory_processor.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR(ProcessorInertialKinematics);
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorInertialKinematics);
} // namespace wolf
