//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file processor_point_feet_nomove.cpp
 *
 *  Created on: Oct 23, 2020
 *      \author: mfourmy
 */

#include "bodydynamics/processor/processor_point_feet_nomove.h"


namespace wolf{


inline ProcessorPointFeetNomove::ProcessorPointFeetNomove(ParamsProcessorPointFeetNomovePtr _params_pfnomove) :
        ProcessorBase("ProcessorPointFeetNomove", 3, _params_pfnomove),
        params_pfnomove_(std::make_shared<ParamsProcessorPointFeetNomove>(*_params_pfnomove)),
        origin_(nullptr)
{
    // Init tracks to nothing
    
    for (int i=0; i < 4; i++){
        tracks_[0] = nullptr;
    }

}

void ProcessorPointFeetNomove::configure(SensorBasePtr _sensor)
{
}



inline void ProcessorPointFeetNomove::processCapture(CaptureBasePtr _capture)
{
    if (!_capture)
    {
        WOLF_ERROR("Received capture is nullptr.");
        return;
    }
    incoming_ = _capture;

    // nothing to do if any of the two buffer is empty
    if(buffer_frame_.empty()){
        WOLF_DEBUG("PInertialKinematic: KF pack buffer empty, time ",  _capture->getTimeStamp());
        return;
    }
    if(buffer_frame_.empty()){
        WOLF_DEBUG("PInertialKinematics: Capture buffer empty, time ",  _capture->getTimeStamp());
        return;
    }

    createFactorConsecutive();
    // processCaptureCreateFactorFar(_capture);

}

inline void ProcessorPointFeetNomove::processKeyFrame(FrameBasePtr _keyframe_ptr)
{
    if (!_keyframe_ptr)
    {
        WOLF_ERROR("Received KF is nullptr.");
        return;
    }
    // nothing to do if any of the two buffer is empty
    if(buffer_frame_.empty()){
        WOLF_DEBUG("ProcessorPointFeetNomove: KF pack buffer empty, time ",  _keyframe_ptr->getTimeStamp());
        return;
    }
    if(buffer_frame_.empty()){
        WOLF_DEBUG("ProcessorPointFeetNomove: Capture buffer empty, time ",  _keyframe_ptr->getTimeStamp());
        return;
    }

    createFactorConsecutive();
    // processKeyFrameCreateFactorFar();

}

void ProcessorPointFeetNomove::processCaptureCreateFactorFar(CaptureBasePtr _capture){
    // Far away factors
    auto cap_curr = std::static_pointer_cast<CapturePointFeetNomove>(_capture);
    auto kin_incontact_current = cap_curr->kin_incontact_;
    for (int i=0; i<4; i++){
        if (!kin_incontact_current.count(i)){
            tracks_[i] = nullptr;
        }
    }
}

void ProcessorPointFeetNomove::processKeyFrameCreateFactorFar(){

    auto sensor_pfnm = std::static_pointer_cast<SensorPointFeetNomove>(getSensor());

    for (auto kf_it: buffer_frame_.getContainer()){
        TimeStamp ts = kf_it.first;
        auto kf = kf_it.second;
        auto cap_it = buffer_capture_.selectIterator(ts, params_->time_tolerance);

        // We have a match between the KF and a capture from the capture buffer
        if (cap_it != buffer_capture_.getContainer().end()){
            auto cap_curr = std::static_pointer_cast<CapturePointFeetNomove>(cap_it->second);
            
            for (auto kin: cap_curr->kin_incontact_){
                // check if each leg in contact has a current active track
                int leg_id = kin.first;
                auto cap_origin = tracks_[leg_id];
                if (!cap_origin){
                    // if not, define current KF as the leg track origin and link the capture to this KF
                    // Error occurs if we try to link several time the same Cap->kf due to the loop on legs -> do only once 
                    if (!cap_curr->getFrame()){
                        cap_curr->link(kf);
                    }
                    tracks_[leg_id] = cap_curr;
                }
                else {
                    if (!cap_curr->getFrame()){
                        cap_curr->link(kf);
                    }
                    // if it has, create a factor
                    auto cap_origin = std::static_pointer_cast<CapturePointFeetNomove>(tracks_[leg_id]);
                    Matrix<double, 14, 1> meas; meas << cap_origin->kin_incontact_[leg_id], kin.second;
                    double dt_kf = cap_curr->getTimeStamp() - cap_origin->getTimeStamp();
                    Matrix3d cov_int = dt_kf*sensor_pfnm->getCovFootNomove();  // integrate zero velocity cov during dt_kf
                    // Matrix3d cov_int = sensor_pfnm->getCovFootNomove();  // integrate zero velocity cov during dt_kf
                    FeatureBasePtr feat = FeatureBase::emplace<FeatureBase>(cap_curr, "PointFeet", meas, cov_int);
                    FactorPointFeetNomovePtr fac = FactorBase::emplace<FactorPointFeetNomove>(feat, feat, cap_origin->getFrame(), nullptr, true);

                    // Zero Altitude factor
                    Vector4d meas_altitude; meas_altitude << kin.second.head<3>(), 0;
                    Matrix1d alt_cov = pow(0.01, 2)*Matrix1d::Identity(); 
                    CaptureBasePtr cap_alt = CaptureBase::emplace<CaptureBase>(kf, "Alt0", cap_curr->getTimeStamp());
                    FeatureBasePtr feat_alt = FeatureBase::emplace<FeatureBase>(cap_alt, "Alt0", meas_altitude, alt_cov);
                    FactorBasePtr fac_alt = FactorBase::emplace<FactorPointFeetAltitude>(feat_alt, feat_alt, nullptr, true);
                }
            }
            buffer_frame_.removeUpTo(cap_curr->getTimeStamp());
            buffer_capture_.removeUpTo(cap_curr->getTimeStamp());
        }
    }
}

void ProcessorPointFeetNomove::createFactorConsecutive(){

    auto sensor_pfnm = std::static_pointer_cast<SensorPointFeetNomove>(getSensor());

    while (buffer_frame_.size() >= 2)
    {
        auto kf1_it = buffer_frame_.getContainer().begin();
        auto kf2_it = std::next(kf1_it);
        TimeStamp t1 = kf1_it->first;
        TimeStamp t2 = kf2_it->first;
        auto cap1_it = buffer_capture_.selectIterator(t1, getTimeTolerance());
        auto cap2_it = buffer_capture_.selectIterator(t2, getTimeTolerance());

        // check that the first 2 KF have corresponding captures in the capture buffer  
        // just quit and assume that you will someday have matching captures
        if ( (cap1_it == buffer_capture_.getContainer().end()) or (cap2_it == buffer_capture_.getContainer().end()) )
        {
            return;
        }
        else
        {
            // store the initial contact/kinematic meas map
            auto cap1 = std::static_pointer_cast<CapturePointFeetNomove>(cap1_it->second);
            auto kin_incontact_from_t1 = cap1->kin_incontact_;

            // then loop over the captures in between cap1 and cap2
            auto cap_it = std::next(cap1_it);
            auto it_after_capt_t2 = std::next(cap2_it);  // used to detect the end of the loop

            // loop throught the captures between t1 and t2
            while (cap_it != it_after_capt_t2){
                // for each new capture, filter the initial contact/kin meas map to track which feet stay in contact throughout
                auto cap_curr = std::static_pointer_cast<CapturePointFeetNomove>(cap_it->second);
                if (!cap_curr){
                    WOLF_ERROR("Wrong capture type in buffer")
                }
                
                // check which feet are currently in contact, if one is missing wrt the previous
                // timestamp, remove it from kin_incontact_from_t1
                auto kin_incontact_current = cap_curr->kin_incontact_;
                for (auto elt_it = kin_incontact_from_t1.cbegin(); elt_it != kin_incontact_from_t1.cend(); /* no increment */){
                    if (kin_incontact_current.find(elt_it->first) == kin_incontact_current.end()){
                        // The foot has lost contact with the ground since t1
                        elt_it = kin_incontact_from_t1.erase(elt_it);  // erase and jump (from stackoverflow)
                    }
                    else{
                        elt_it++;
                    }
                }
                cap_it++;
            }

            if (!kin_incontact_from_t1.empty())
            {
                auto cap2 = std::static_pointer_cast<CapturePointFeetNomove>(cap2_it->second);
                // link (kind of arbitrarily) factor from KF1 to KF2 with a feature and capture on KF2 to mimic capture motion
                // however, this capture solely does not contain enough information to recreate the factor
                cap2->link(kf2_it->second);
                auto kin_incontact_t2 = cap2->kin_incontact_;

                for (auto kin_pair1: kin_incontact_from_t1){
                    auto kin_pair2_it = kin_incontact_t2.find(kin_pair1.first);

                    Matrix<double, 14, 1> meas; meas << kin_pair1.second, kin_pair2_it->second;
                    double dt_kf = cap2->getTimeStamp() - cap1->getTimeStamp();
                    Matrix3d cov_int = dt_kf*sensor_pfnm->getCovFootNomove();  // integrate zero velocity cov during dt_kf
                    FeatureBasePtr feat = FeatureBase::emplace<FeatureBase>(cap2, "PointFeet", meas, cov_int);
                    FactorPointFeetNomovePtr fac = FactorBase::emplace<FactorPointFeetNomove>(feat, feat, kf1_it->second, nullptr, true);

                    // Zero Altitude factor
                    Vector4d meas_altitude; meas_altitude << kin_pair2_it->second.head<3>(), 0;
                    Matrix1d alt_cov = pow(0.01, 2)*Matrix1d::Identity(); 
                    CaptureBasePtr cap_alt = CaptureBase::emplace<CaptureBase>(kf2_it->second, "Alt0", cap2->getTimeStamp());
                    FeatureBasePtr feat_alt = FeatureBase::emplace<FeatureBase>(cap_alt, "Alt0", meas_altitude, alt_cov);
                    FactorBasePtr fac_alt = FactorBase::emplace<FactorPointFeetAltitude>(feat_alt, feat_alt, nullptr, true);
                }
            }

            // Once the factors are created, remove kf1 and all the captures until ts 2 NOT INCLUDING the one at ts 2 since we need it for the next
            // Note: erase by range does not include the end range iterator
            buffer_frame_.getContainer().erase(buffer_frame_.getContainer().begin(), kf2_it);  // !! works only if getContainer returns a non const reference
            buffer_capture_.getContainer().erase(buffer_capture_.getContainer().begin(), cap2_it);
        }
    }
}




inline bool ProcessorPointFeetNomove::voteForKeyFrame() const
{
    return false;
}



} /* namespace wolf */

// Register in the FactoryProcessor
#include "core/processor/factory_processor.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR(ProcessorPointFeetNomove);
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorPointFeetNomove);
} // namespace wolf
