//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

#include "bodydynamics/factor/factor_force_torque_inertial_dynamics.h"
#include "bodydynamics/factor/factor_angular_momentum.h"
#include "bodydynamics/processor/processor_force_torque_inertial_dynamics.h"
#include "bodydynamics/sensor/sensor_force_torque_inertial.h"
#include "bodydynamics/internal/config.h"
#include "bodydynamics/capture/capture_force_torque_inertial.h"

#include <core/sensor/factory_sensor.h>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>
#include <core/utils/params_server.h>
#include <core/yaml/parser_yaml.h>
#include <core/state_block/state_block_derived.h>
#include <core/state_block/factory_state_block.h>
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/math/rotations.h>
#include <core/tree_manager/tree_manager_sliding_window.h>
#include <core/capture/capture_pose.h>


#include <Eigen/Dense>
#include <Eigen/Geometry>

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>

using namespace wolf;
using namespace bodydynamics::fti;

WOLF_REGISTER_STATEBLOCK_WITH_KEY(L, StateVector3d);

class Test_SimulationProblemForceTorqueInertialDynamics_yaml : public testing::Test
{
  public:
    ProblemPtr                              P;
    SolverCeresPtr                          solver;
    
    // FTI sensor
    SensorForceTorqueInertialPtr            S;
    ProcessorForceTorqueInertialDynamicsPtr p;
    Vector6d                                bias_true;
    Vector3d                                cdm_true;
    Vector3d                                inertia_true;
    double                                  mass_true;
    Matrix<double, 12, 1>                   data_fti;
    
    // Pose sensor
    SensorBasePtr                           SP;
    Vector7d                                data_pose;

    double                                  dt;

    // Reading csv data_fti
    std::vector<double>                     time_stamp;
    std::vector<Vector3d>                   position, vlin, vang, force, torque, a_meas;
    std::vector<Quaterniond>                quaternion;

  protected:
    void extractAndCompleteData(const std::string& file_path_name)
    {
        Vector3d      position_i, vlin_i, vang_i, force_i, torque_i, a_meas_i;
        std::ifstream data_simulation;
        string        data_file = file_path_name;
        data_simulation.open(data_file, std::ios::in);
        string line_data;
        char   delimiter = ',';
        std::getline(data_simulation, line_data);

        int counter = 0;

        while (std::getline(data_simulation, line_data))
        {
            std::stringstream data_str(line_data);

            // time stamp
            string string_time_stamp;
            std::getline(data_str, string_time_stamp, delimiter);
            time_stamp.push_back(std::stod(string_time_stamp));

            // position
            string string_pos_x;
            string string_pos_y;
            string string_pos_z;

            std::getline(data_str, string_pos_x, delimiter);
            std::getline(data_str, string_pos_y, delimiter);
            std::getline(data_str, string_pos_z, delimiter);

            position_i(0) = std::stod(string_pos_x);
            position_i(1) = std::stod(string_pos_y);
            position_i(2) = std::stod(string_pos_z);

            position.push_back(position_i);

            // quaternion
            string string_quat_x;
            string string_quat_y;
            string string_quat_z;
            string string_quat_w;

            std::getline(data_str, string_quat_x, delimiter);
            std::getline(data_str, string_quat_y, delimiter);
            std::getline(data_str, string_quat_z, delimiter);
            std::getline(data_str, string_quat_w, delimiter);

            Quaterniond quaternion_i(std::stod(string_quat_w),
                                     std::stod(string_quat_x),
                                     std::stod(string_quat_y),
                                     std::stod(string_quat_z));

            quaternion.push_back(quaternion_i);

            // linear velocity
            string string_vlin_x;
            string string_vlin_y;
            string string_vlin_z;

            std::getline(data_str, string_vlin_x, delimiter);
            std::getline(data_str, string_vlin_y, delimiter);
            std::getline(data_str, string_vlin_z, delimiter);

            vlin_i(0) = std::stod(string_vlin_x);
            vlin_i(1) = std::stod(string_vlin_y);
            vlin_i(2) = std::stod(string_vlin_z);

            vlin.push_back(vlin_i);

            // angular velocity
            string string_vang_x;
            string string_vang_y;
            string string_vang_z;

            std::getline(data_str, string_vang_x, delimiter);
            std::getline(data_str, string_vang_y, delimiter);
            std::getline(data_str, string_vang_z, delimiter);

            vang_i(0) = std::stod(string_vang_x);
            vang_i(1) = std::stod(string_vang_y);
            vang_i(2) = std::stod(string_vang_z);

            vang.push_back(vang_i);

            // force
            string string_force_x;
            string string_force_y;
            string string_force_z;

            std::getline(data_str, string_force_x, delimiter);
            std::getline(data_str, string_force_y, delimiter);
            std::getline(data_str, string_force_z, delimiter);

            force_i(0) = std::stod(string_force_x);
            force_i(1) = std::stod(string_force_y);
            force_i(2) = std::stod(string_force_z);

            force.push_back(force_i);

            // torque
            string string_torque_x;
            string string_torque_y;
            string string_torque_z;

            std::getline(data_str, string_torque_x, delimiter);
            std::getline(data_str, string_torque_y, delimiter);
            std::getline(data_str, string_torque_z, delimiter);

            torque_i(0) = std::stod(string_torque_x);
            torque_i(1) = std::stod(string_torque_y);
            torque_i(2) = std::stod(string_torque_z);

            torque.push_back(torque_i);

            // acceleration -- need to compute from other data
            a_meas_i = force_i/mass_true;

            a_meas.push_back(a_meas_i);
            
            counter++;
        }
    }

    void SetUp() override
    {
        bias_true    = Vector6d::Zero();
        cdm_true     = {0, 0, 0};
        inertia_true = {0.0134943, 0.0141622, 0.0237319};
        mass_true    = 1.9828;

        std::string wolf_bodydynamics_root = _WOLF_BODYDYNAMICS_ROOT_DIR;

        extractAndCompleteData(wolf_bodydynamics_root + "/test/dades_simulacio_pep.csv");
        

        ParserYaml   parser = ParserYaml("problem_force_torque_inertial_dynamics_simulation_test.yaml",
                                       wolf_bodydynamics_root + "/test/yaml/");
        ParamsServer server = ParamsServer(parser.getParams());
        P                   = Problem::autoSetup(server);

        solver       = std::make_shared<SolverCeres>(P);
        auto options = solver->getSolverOptions();
        // solver->getSolverOptions().max_num_iterations               = 1e4;
        // solver->getSolverOptions().function_tolerance               = 1e-15;
        // solver->getSolverOptions().gradient_tolerance               = 1e-15;
        // solver->getSolverOptions().parameter_tolerance              = 1e-15;

        // FTI sensor and processor
        S = std::static_pointer_cast<SensorForceTorqueInertial>(P->findSensor("sensor FTI"));
        p = std::static_pointer_cast<ProcessorForceTorqueInertialDynamics>(S->getProcessorList().front());

        // Pose sensor
        SP = P->findSensor("sensor Pose");
    }
};

TEST_F(Test_SimulationProblemForceTorqueInertialDynamics_yaml, force_registration)
{
    FactorySensor::registerCreator("SensorForceTorqueInertial", SensorForceTorqueInertial::create);
}

//this test checks the pre-integration evolution along the time
TEST_F(Test_SimulationProblemForceTorqueInertialDynamics_yaml, preintegration_and_csv)
{
    std::string wolf_bodydynamics_root = _WOLF_BODYDYNAMICS_ROOT_DIR;
    // creem un nou arxiu CSV per imprimir els estats que estem rebent ara
    std::fstream fout;
    fout.open(wolf_bodydynamics_root + "/test/dades_simulacio_pep_estimador.csv",
                std::fstream::out | std::fstream::trunc);

    S->getStateBlock('I')->setState(bias_true);
    S->getStateBlock('C')->setState(cdm_true);
    S->getStateBlock('i')->setState(inertia_true);
    S->getStateBlock('m')->setState(Vector1d(mass_true));

    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->fix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->fix();
    S->setStateBlockDynamic('I', false);

    //Writing the first lines to know how the data will be distributed in the csv
    fout << "time stamp"
         << ","
         << "position_x_true"
         << ","
         << "position_y_true"
         << ","
         << "position_z_true"
         << ","
         << "quaternion_x_true"
         << ","
         << "quaternion_y_true"
         << ","
         << "quaternion_z_true"
         << ","
         << "quaternion_w_true"
         << ","
         << "position_x_est"
         << ","
         << "position_y_est"
         << ","
         << "position_z_est"
         << ","
         << "quaternion_x_est"
         << ","
         << "quaternion_y_est"
         << ","
         << "quaternion_z_est"
         << ","
         << "quaternion_w_est"
         << "\n";

    // Force processor to make a KF at t=0
    CaptureMotionPtr C0 =
        std::make_shared<CaptureForceTorqueInertial>(0, S, VectorXd::Zero(12), MatrixXd::Identity(12, 12), nullptr);
    C0->process();

    int i_init = 0;

    // fix measured position and orientation
    p->getOrigin()->getFrame()->getStateBlock('P')->fix();
    p->getOrigin()->getFrame()->getStateBlock('P')->setState(position[i_init]);
    p->getOrigin()->getFrame()->getStateBlock('O')->fix();
    p->getOrigin()->getFrame()->getStateBlock('O')->setState(quaternion[i_init].coeffs());
    p->getOrigin()->getFrame()->getStateBlock('V')->fix();
    p->getOrigin()->getFrame()->getStateBlock('V')->setState(quaternion[i_init].conjugate() * vlin[i_init]);
    Vector3d ang_mom_init(
        vang[i_init](0) * inertia_true(0), vang[i_init](1) * inertia_true(1), vang[i_init](2) * inertia_true(2));
    p->getOrigin()->getFrame()->getStateBlock('L')->fix();
    p->getOrigin()->getFrame()->getStateBlock('L')->setState(ang_mom_init);


    for (int i = i_init + 1; i < time_stamp.size(); i++)  // start with second data_fti
    {
        // SIMULATOR

        TimeStamp t = time_stamp[i];

        // Data
        data_fti << a_meas[i], vang[i], force[i], torque[i];

        // ESTIMATOR

        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data_fti, S->getNoiseCov(), nullptr);
        C->process();

        // results printed in the csv in the order established before
        Vector3d position_est(p->getLast()->getFrame()->getStateBlock('P')->getState());
        Vector4d quaternion_coeffs_est(p->getLast()->getFrame()->getStateBlock('O')->getState());

        fout << time_stamp[i] << "," << position[i](0) << "," << position[i](1) << "," << position[i](2) << ","
             << quaternion[i].coeffs()(0) << "," << quaternion[i].coeffs()(1) << "," << quaternion[i].coeffs()(2)
             << "," << quaternion[i].coeffs()(3) << "," << position_est(0) << "," << position_est(1) << ","
             << position_est(2) << "," << quaternion_coeffs_est(0) << "," << quaternion_coeffs_est(1) << ","
             << quaternion_coeffs_est(2) << "," << quaternion_coeffs_est(3)
             << "\n";

    }
    fout.close();
}

TEST_F(Test_SimulationProblemForceTorqueInertialDynamics_yaml, simulation_online)
{
    // Calibration params
    Vector6d bias_guess    = S->getStateBlock('I')->getState();
    Vector3d cdm_guess     = S->getStateBlock('C')->getState();
    Vector3d inertia_guess = S->getStateBlock('i')->getState();
    double   mass_guess    = S->getStateBlock('m')->getState()(0);

    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->unfix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->unfix();
    S->getStateBlock('m')->unfix();
    S->setStateBlockDynamic('I', false);


    // add regularization, uncomment if the parameter is not fixed

    S->addPriorParameter('I',                   // cdm
                         bias_guess,            // cdm
                         0.1*0.1*Matrix6d::Identity());  // cov

    S->addPriorParameter('C',                   // cdm
                         cdm_guess,             // cdm
                         0.1*0.1*Matrix3d::Identity()); // cov

    S->addPriorParameter('i',                   // inertia
                         inertia_guess,         // inertia
                         0.01*0.01*Matrix3d::Identity()); // cov

    S->addPriorParameter('m',                               // mass
                         Vector1d(mass_guess),              // mass
                         0.1 * 0.1 * Matrix1d::Identity()); // cov

    std::string report;

    // Force processor to make a KF at t=0
    CaptureMotionPtr C0 =
        std::make_shared<CaptureForceTorqueInertial>(0, S, VectorXd::Zero(12), MatrixXd::Identity(12, 12), nullptr);
    C0->process();

    int i_init = 0;

    // Pose
    data_pose << position[i_init] , quaternion[i_init].coeffs();
    CaptureBasePtr CP = std::make_shared<CapturePose>(time_stamp[i_init], SP, data_pose, SP->getNoiseCov());
    CP->process();

    unsigned int last_kf_id = p->getOrigin()->getFrame()->id();

    for (int i = i_init + 1; i < time_stamp.size(); i++)  // start with second data
    {
        // SIMULATOR

        TimeStamp t = time_stamp[i];

        // Data
        data_fti  << a_meas[i], vang[i], force[i], torque[i];
        data_pose << position[i] , quaternion[i].coeffs();

        // ESTIMATOR

        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data_fti, S->getNoiseCov(), nullptr);
        C->process();

        CaptureBasePtr CP = std::make_shared<CapturePose>(t, SP, data_pose, SP->getNoiseCov());
        CP->process();

        // check if new KF and solve
        if (last_kf_id != p->getOrigin()->getFrame()->id())
        {
            last_kf_id = p->getOrigin()->getFrame()->id();

            // solve!
            report = solver->solve(wolf::SolverManager::ReportVerbosity::BRIEF);
        
            // results of this iteration
            WOLF_INFO("Time                    : ", t, " s.");
            WOLF_INFO(report);
            WOLF_INFO("Estimated bias          : ", S->getStateBlock('I')->getState().transpose(), " m/s2 - rad/s.");
            WOLF_INFO("Estimated inertia       : ", S->getStateBlock('i')->getState().transpose(), " m^2 Kg.");
            WOLF_INFO("Estimated center of mass: ", S->getStateBlock('C')->getState().transpose(), " m.");
            WOLF_INFO("Estimated mass          : ", S->getStateBlock('m')->getState()(0), " Kg.");
            WOLF_INFO("-----------------------------");

        }
    }

    // FINAL RESULTS
    WOLF_INFO("True bias             * : ", bias_true.transpose(), " m/s2 - rad/s.");
    WOLF_INFO("Guess bias              : ", bias_guess.transpose(), " m/s2 - rad/s.");
    WOLF_INFO("Estimated bias          : ", S->getStateBlock('I')->getState().transpose(), " m/s2 - rad/s.");
    WOLF_INFO("True center of mass   * : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess center of mass    : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated center of mass: ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("True inertia          * : ", inertia_true.transpose(), " m^2 Kg.");
    WOLF_INFO("Guess inertia           : ", inertia_guess.transpose(), " m^2 Kg.");
    WOLF_INFO("Estimated inertia       : ", S->getStateBlock('i')->getState().transpose(), " m^2 Kg.");
    WOLF_INFO("True mass             * : ", mass_true, " Kg.");
    WOLF_INFO("Guess mass              : ", mass_guess, " Kg.");
    WOLF_INFO("Estimated mass          : ", S->getStateBlock('m')->getState()(0), " Kg.");
    WOLF_INFO("-----------------------------");

    ASSERT_MATRIX_APPROX(bias_true   , S->getStateBlock('I')->getState(), 0.11);
    ASSERT_MATRIX_APPROX(cdm_true    , S->getStateBlock('C')->getState(), 1e-2);
    ASSERT_MATRIX_APPROX(inertia_true, S->getStateBlock('i')->getState(), 1e-2);
    ASSERT_NEAR         (mass_true   , S->getStateBlock('m')->getState()(0), 2e-2);
}

TEST_F(Test_SimulationProblemForceTorqueInertialDynamics_yaml, simulation_batch)
{
    // Calibration params
    Vector6d bias_guess    = S->getStateBlock('I')->getState();
    Vector3d cdm_guess     = S->getStateBlock('C')->getState();
    Vector3d inertia_guess = S->getStateBlock('i')->getState();
    double   mass_guess    = S->getStateBlock('m')->getState()(0);

    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->unfix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->unfix();
    S->getStateBlock('m')->unfix();
    S->setStateBlockDynamic('I', false);


    // // add regularization, uncomment if the parameter is not fixed

    // S->addPriorParameter('I',                                // bias
    //                      bias_guess,                         // bias
    //                      0.1 * 0.1 * Matrix3d::Identity());  // cov

    // S->addPriorParameter('C',                                  // cdm
    //                      cdm_guess,                            // cdm
    //                      0.01 * 0.01 * Matrix3d::Identity());  // cov

    // S->addPriorParameter('i',                                  // inertia
    //                      inertia_guess,                        // inertia
    //                      0.01 * 0.01 * Matrix3d::Identity());  // cov

    // S->addPriorParameter('m',                                // mass
    //                      Vector1d(mass_guess),               // mass
    //                      0.01 * 0.01 * Matrix1d::Identity());  // cov

    int i_init = 0;
    
    // Force FTI processor to make a KF at t=0
    CaptureMotionPtr C0 =
        std::make_shared<CaptureForceTorqueInertial>(0, S, VectorXd::Zero(12), S->getNoiseCov(), nullptr);
    C0->process();

    // Pose
    data_pose << position[i_init] , quaternion[i_init].coeffs();
    CaptureBasePtr CP = std::make_shared<CapturePose>(time_stamp[i_init], SP, data_pose, SP->getNoiseCov());
    CP->process();

    for (int i = i_init + 1; i < time_stamp.size(); i++)  // start with second data
    {
        // SIMULATOR
        TimeStamp t = time_stamp[i];

        // Data
        data_fti  << a_meas[i], vang[i], force[i], torque[i];
        data_pose << position[i] , quaternion[i].coeffs();

        // ESTIMATOR

        // fti
        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data_fti, S->getNoiseCov(), nullptr);
        C->process();

        // Pose
        CaptureBasePtr CP = std::make_shared<CapturePose>(t, SP, data_pose, SP->getNoiseCov());
        CP->process();
    }

    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::BRIEF);


    // FINAL RESULTS
    WOLF_INFO("-----------------------------");
    WOLF_INFO(report);
    WOLF_INFO("True bias             * : ", bias_true.transpose(), " m/s2 - rad/s.");
    WOLF_INFO("Guess bias              : ", bias_guess.transpose(), " m/s2 - rad/s.");
    WOLF_INFO("Estimated bias          : ", S->getStateBlock('I')->getState().transpose(), " m/s2 - rad/s.");
    WOLF_INFO("True center of mass   * : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess center of mass    : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated center of mass: ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("True inertia          * : ", inertia_true.transpose(), " m^2 Kg.");
    WOLF_INFO("Guess inertia           : ", inertia_guess.transpose(), " m^2 Kg.");
    WOLF_INFO("Estimated inertia       : ", S->getStateBlock('i')->getState().transpose(), " m^2 Kg.");
    WOLF_INFO("True mass             * : ", mass_true, " Kg.");
    WOLF_INFO("Guess mass              : ", mass_guess, " Kg.");
    WOLF_INFO("Estimated mass          : ", S->getStateBlock('m')->getState()(0), " Kg.");
    WOLF_INFO("-----------------------------");

    ASSERT_MATRIX_APPROX(bias_true   , S->getStateBlock('I')->getState(), 0.21);
    ASSERT_MATRIX_APPROX(cdm_true    , S->getStateBlock('C')->getState(), 2e-3);
    ASSERT_MATRIX_APPROX(inertia_true, S->getStateBlock('i')->getState(), 1e-2);
    ASSERT_NEAR         (mass_true   , S->getStateBlock('m')->getState()(0), 4e-2);
}

int main(int argc, char** argv)
{
    testing::InitGoogleTest(&argc, argv);
    // ::testing::GTEST_FLAG(filter) =
        // "Test_SimulationProblemForceTorqueInertialDynamics_yaml.preintegration_and_csv";
    ::testing::GTEST_FLAG(filter) =
        "Test_SimulationProblemForceTorqueInertialDynamics_yaml.simulation*";
        
    return RUN_ALL_TESTS();
}
